package PateLib;

import java.util.*;
import java.awt.*;

/*
---------------------------------------------------------------------------
File	    :		Grammar.java
Package	    :		PateLib  Version 1.0
Author	    :		Jason Robert Salemme
Programmed  :		8 Aug 96	
Modified    :
---------------------------------------------------------------------------
 *
 * 
 * Susan H. Rodger, Anna O. Bilska (Anya), Jason R. Salemme
 * Computer Science Department
 * Duke University
 * August 1996
 * Supported by National Science Foundation DUE-9596002 and DUE-9555084.
 *
 * Copyright (c) 1996
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the author.  The name of the author may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 ****************************************************************************/
/**
 * This is a generic class that is used to store a grammar.  Actually, it
 * stores the productions of the grammar.  You are able to add and delete 
 * productions as well check them to make sure that they are syntatically 
 * correct (you are not able to add a production unless it is in
 * correct syntax).  A host of other methods allow you to manipulate 
 * them.
 *
 * <PRE>
 * Note:    All variable are capital letters with.  Two variables can exist in
 * different forms.  The variable B can be by itself or modified with a
 * terminal in the form B(a).  The variable D can also be by itself or
 * modified with an integer in the form D(#)
 * </PRE>
 * @author Jason R. Salemme
 * @version 1.0, 8 Aug 96
*/
public class Grammar implements Cloneable
{   
/**
 * The start variable for the grammar
 * @see setStartSymbol(String s)
 * @see getStartSymbol()
 * @see HasStartProd()
 */
    private char myStart;

    //If you want to add characters for acceptable terminals, be my guest
/**
 * The set of additional characters (other than lower case letter) that are
 * allowed as valid terminals.
 * @see Patelib.Grammar.IsOther(char c)
 */
    final static char[] others = {'+', '-', '/', '*', '%', '.', '[', ']',
				  '<', '>'};
    //If you want to change the look of the arrow, be my guest.
/**
 * The definition of the "arrow" that seperates the left side from the right 
 * side of the production.
 */
    final static String arrow = "->";

/**
 * Allocates a new instance of class Grammar.
 */
    public Grammar()  //Constructor
    {
	myVariables = new Vector();
	myTerminals = new Vector();
	myStart = 'S';
	myProdsLeft = new Vector();
	myProdsRight = new Vector();
	myNumVars = 0;
	myNumTerms = 0;
	myNumProds = 0;
	myV = 0;
	myT = 0;
	myP = 0;
    }

/**
 * Creates a clone of the grammar.
 */
    public Object clone()
    {
	try
	{
	    Grammar g = (Grammar)super.clone();
	    g.myVariables = (Vector)myVariables.clone();
	    g.myTerminals = (Vector)myTerminals.clone();
	    g.myProdsLeft = (Vector)myProdsLeft.clone();
	    g.myProdsRight = (Vector)myProdsRight.clone();
	    return g;

	}catch (CloneNotSupportedException e)
	{   //This should not happen since we are Cloneable
	    return null;
	}
    }

/**
 * Determines if a grammar is "equal to" this one.
 */
    public boolean equals(Grammar g)
    //postcond:  returns true is the two grammars are the same, or false
    {
	if(myNumProds != g.myNumProds)
	    return false;

	for(int k = 0; k < myNumProds; k++)
	{
	    if(!g.IsAProd((String)myProdsLeft.elementAt(k) + arrow +
		(String)myProdsRight.elementAt(k)))
		return false;
	}
	return true;
    }

/**
 * Determines if a production is syntatically correct.
 * @param p the string representing the production
 */
    public boolean GoodProduction(String p){
    //postcond: Returns true if string p is a syntatically correct production
    //		and adds the production to the grammar
	char c;
	int skip = 0;

	//if less than 3 characters is can not be a rule!
	if(p.length() < 3)
	    return false;
	//if it does not have an arrow (or more than one) do not add it
	if(!containsArrow(p))
	    return false;
	//Check the rest of the string
	for(int k = 0; k < p.length(); k++){
	    c = p.charAt(k);
	    //Check to see if acceptable characters
	    if(!Character.isUpperCase(c) && 
		    !Character.isLowerCase(c) && 
		    !Character.isDigit(c) && !IsOther(c) && 
		    c != Params.lambdaCharValue)
            {
		return false;
	    }
	    if(c == Params.lambdaChar && k < myArrowPosition)
		return false;
	    //Check to see if the B var is in B or B(a) form
	    if(c == 'B'){
		skip = ValidBVar(k, p);
		    if(skip == -1)
			return false;
		    k += skip;
	    }

	    //Check to see if the D var is in D or D(#) form
	    if(c == 'D'){
		skip = ValidDVar(k, p);
		    if(skip == -1)
			return false;
		    k += skip;
	    }
	}
	//If loop fully iterated		 
	return true;
    }

/**
 * Adds a valid production to the grammar.
 * @param p production to be added
 */
    public void addProduction(String p)  //Adds a production, variable, etc
    //precond:	 the produciton (STRING P) should be valid
    //postcond:  adds a production to the grammar
    {
        //remove spaces
        String prod = new String(p);
        prod = prod.trim();
        for(int x = 0; x < prod.length(); ++x)
        {
            if(prod.charAt(x) == ' ')
            {
               prod = prod.substring(0, x) + prod.substring(x + 1, prod.length());
               --x;
            }
        }
   
	//Double checking for validity (bozo proofing)
	if(GoodProduction(prod))
        {
	    //Get the left side
	    String word = prod.substring(0, myArrowPosition - 1);
	    AddTermsNVars(word);
	    myProdsLeft.addElement(word);   //add left side of production
	    //get the right side
	    word = prod.substring(myArrowPosition + 1, prod.length());
            if(word.equals(""))
               word = Params.lambdaChar + "";
	    AddTermsNVars(word);
	    myProdsRight.addElement(word);  //add right side of production
	    myNumProds++;		    //add 1 to Prods counter
	}
	else
        {
	    System.out.println(prod);
	    System.out.println("Bummer... Production not added!");
	}
    }

/**
 * Removes the production that is currently being "pointed" to within the class
 */
    public void RemoveCurrentProd()  //Removes a production
    {
	myProdsLeft.removeElementAt(myP);
	myProdsRight.removeElementAt(myP);
	myNumProds--;
	myP--;

	//There is a more efficient way to do this but for the sake
	//of time I did the first thing that came to mind.
	//This updates the myVariables and myTerminals in case one of
	//them does not exist anymore after the production was removed.
	myTerminals = new Vector();
	myNumTerms = 0;
	myT = 0;
	myVariables = new Vector();
	myNumVars = 0;
	myV = 0;
	for (int k = 0; k < myNumProds; k++){
	    AddTermsNVars((String)myProdsLeft.elementAt(k) + 
			  (String)myProdsRight.elementAt(k));
	}
    }

/**
 * Returns the number of spaces to skip for the string index to point at the
 * end of the variable.  The variable <i>MUST</i> be a B variable.
 * @param k index to the B variable
 * @param p the left or right side, or the full production
 */
    public int ValidBVar(int k, String p){
    //postcond:  returns # spaces to skip if B var is in form B(a) or B
    //		 returns -1 if it is not a correct B variable 
	//Check at the end of the left or right side of the prod
	if(k == p.length()-1 || p.charAt(k+1) == '=')
	    return 0;

	if(p.charAt(k+1) == ')')    //Check wrong parenthesis
	    return -1;
	
	if(p.charAt(k+1) == '('){
	    //Make sure there are enough character left for valid B
	    if(k > p.length()-4){
		return -1;
	    }
	    if(!(Character.isLowerCase(p.charAt(k+2)) ||
		    IsOther(p.charAt(k+2))) || 
		    p.charAt(k+3) != ')')
		return -1;
	    //If made it here return skip 3 spaces to put at end
	    //of the B(a) variable
	    return 3;
	}
	
	//If made it here skip 0 spaces
	return 0;
    }

/**
 * Returns the number of spaces to skip for the string index to point at the
 * end of the variable.  The variable <i>MUST</i> be a D variable.
 * @param k index to the D variable
 * @param p the left or right side, or the full production
 */
    public int ValidDVar(int k, String p){
    //postcond:  returns # spaces to skip if D var is in form D(#) or B
    //		 returns -1 if it is not a correct D varaible
	int counter = 0;
	int j;

	//Check at the end of the left or right side of the prod
	if(k == p.length()-1 || p.charAt(k+1) == '=')
	    return 0;

	if(p.charAt(k+1) == ')')    //Check wrong parenthesis
	    return -1;
	
	if(p.charAt(k+1) == '('){
	    //Make sure there are enough character left for valid D
	    if(k > p.length()-4)
		return -1;
	    counter = 2;
	    for(j = k+2; j < p.length(); j++){
		if(p.charAt(j) == ')')
		    break;
		else if(!Character.isDigit(p.charAt(j)))
		    return -1;
		counter++;
	    }
	    //Make sure there was a number inside
	    if(j == k+2 || j == p.length())
		return -1;
	}
	
	//If made it here skip counter spaces
	return counter;
    }

/**
 * Indicates whether the character is in the list of other acceptable
 * characters for terminals.
 * @param c the character in question
 */
    public boolean IsOther(char c)
    {
	for(int k = 0; k < others.length; k++){
	    if(others[k] == c)
		return true;
	}
	return false; 
    }
	
/**
 * Indicates whether the production is in this grammar.
 * @param prod the production in question
 */
    public boolean IsAProd(String prod){
	for(int k = 0; k < myNumProds; k++){
	    if(prod.equals((String)myProdsLeft.elementAt(k) + arrow + 
			(String)myProdsRight.elementAt(k)))
		return true;
	}
	return false;
    }

/**
 * Indicates if the variable exists in the grammar.
 * @param v the variable in question
 */
    public boolean IsInVars(String v){
	return myVariables.indexOf(v) != -1;
    }

/**
 * Indicates if the terminal exists in the grammar.
 * @param t the terminal in question
 */
    public boolean IsInTerms(String t){
	return myTerminals.indexOf(t) != -1;
    }

/**
 * Set the class variable index to the first variable it has in memory.
 */
    public void FirstVar(){
	myV = 0;
    }    

/**
 * Set the class production index to the first production in the grammar.
 */
    public void FirstProd(){
	myP = 0;
    }    

/**
 * Increment the variable index to the next variable.
 */
    public void NextVar(){
	myV++;
    }

/**
 * Increment the production to the next production.
 */
    public void NextProd(){
	myP++;
    }

/**
 * Returns the current variable.
 */
    public String CurrentVar(){
	return (String)myVariables.elementAt(myV);
    }

/**
 * Returns the current left side of the production.
 */
    public String CurrentLeftProd(){
	return (String)myProdsLeft.elementAt(myP);
    }

/**
 * Returns the current right side of the production.
 */
    public String CurrentRightProd(){
	return (String)myProdsRight.elementAt(myP);
    }

/**
 * Returns the current production.
 */
    public String CurrentProd(){
	return (String)myProdsLeft.elementAt(myP) + arrow +
	       (String)myProdsRight.elementAt(myP);
    }

/**
 * Indicates if the variable index is at the last variable.
 */
    public boolean VarsDone(){
	return (myV == myNumVars);
    }

/**
 * Indicates if the production index is at the last production.
 */
    public boolean ProdsDone(){
	return (myP == myNumProds);
    }

/**
 * Returns the number of production in the grammar.
 */
    public int NumProductions(){
	return myNumProds;
    }

/**
 * Returns the number of productions in the grammar
 */
    public String getStartSymbol(){
	return "" + myStart;
    }
   
/**
 * Sets the start symbol to a new value.
 * @param c the character for the new start symbol
 */
    public void setStartSymbol(char c){
	myStart = c;
    }

/**
 * Indicates if the grammar has a starting production.
 */
    public boolean HasStartProd(){
	return myProdsLeft.indexOf("" + myStart) != -1;
    }	

    public void PrintProductions(){
    //postcondition:  prints out all productions
	for(int k = 0; k < myNumProds; k++){
	    System.out.println((String)myProdsLeft.elementAt(k) +
		arrow + (String)myProdsRight.elementAt(k));
	}
    }

/**
 * Returns a vector containing the variables in the grammar.
 */
    public Vector getVariables(){
    //returns a vector of the variables
	return (Vector)myVariables.clone();
    }

/**
 * Indicates if the grammar is unrestricted or not.
 */
    public boolean IsUnrestricted()
    {
	//returns true is the grammar is unrestricted
	String str;
	for(int k = 0; k < myNumProds; k++)
	{
	    str = (String)myProdsLeft.elementAt(k);
	    
	    if(getEnd(str, 0) < str.length())
		return true;
	}
	return false;
    }
  
/**
 * Returns the index to the proceeding terminal or variable after a B or D
 * variable.
 * @param s the string representing the full/partial production.
 * @param k the index to the B or D variable
 */
    public int getEnd(String s, int k){
    //precond:	k must be the index to a variable in a valid production
	int j;
	if((s.charAt(k) == 'B' || s.charAt(k) == 'D') &&
	    k <= s.length() - 4 && s.charAt(k+1) == '('){
	    for(j = k+3; j < s.length(); j++){
		if(s.charAt(j) == ')')
		    return j + 1;
	    }
        }
	//For varaibles that are one character long
	return k+1;
    }

/**
 * Indicates if the current production terminates through terminals.  This
 * does not determine if the production <i>derives</i> terminals.
 */
    public boolean Terminates(){
    //precondition:  the production is already a good production
	String s = (String)myProdsRight.elementAt(myP);
	for(int k = 0; k < s.length(); k++){
	    if(Character.isUpperCase(s.charAt(k)))
		return false;
	}
	return true;
    }

    private void addVariable(String c){
    //postcondition: adds the variable if not already in it
	if(!IsInVars(c)){
	    myVariables.addElement(c);
	    myNumVars++;
	}
    }

    private void addTerminal(String c){
    //postcondition:  adds the terminal if not already in it
	if(!IsInTerms(c)){
	    myTerminals.addElement(c);	
	    myNumTerms++;
	}
    }

    private void AddTermsNVars(String s){
    //precondition: string contains only terminals and variables
    //postcondition: adds terminals and variables
	String c;
	char cc;
	int skip = 0;

	for(int k = 0; k < s.length(); k++){
	    c = "" + s.charAt(k);
	    cc = s.charAt(k);

	    //if a variable add it
	    if(Character.isUpperCase(cc)){
		if(cc == 'B'){
		    skip = ValidBVar(k, s);
		    addVariable(s.substring(k, k + skip+1));
		    k += skip;
		}
		else if(cc == 'D'){
		    skip = ValidDVar(k, s);
		    addVariable(s.substring(k, k + skip+1));
		    k += skip;
		}
		else
		    addVariable(s.substring(k, k+1));
	    }
	    //if a terminal add it
	    else if(Character.isLowerCase(cc) || IsOther(cc))
		addTerminal(c);    
	}
    }

    private boolean containsArrow(String s){
    //postcond:  returns true if the string contains the string "->" once.
	//length -1 because there doesn't have to be something on the right side
	//of the production...
	int count = 0;
	for(int k = 0; k < s.length() - 1; k++)
        {
	    if(arrow.equals(s.substring(k, k + arrow.length())))
            {
		count++;
		myArrowPosition = k+1;	//end of arrow;
	    }
	}
	return count == 1;
    }
  
    public String RightProdAt(int i)
    // returns the right production at position i
    {
	if (i>myNumProds || i<0)
	    return null;
	return (String)myProdsRight.elementAt(i);
    }
		
    public String LeftProdAt(int i)
    // returns the left production at position i
    {
	if (i>myNumProds || i<0)
	    return null;
	return (String)myProdsLeft.elementAt(i);
    }

    public boolean IsPossibleTerminal(char c)
    // postcondition: returns ture if 'c' could be
    // a terminal (but not necessarily in this grammar)
    {
	if (Character.isLowerCase(c))
	    return true;
	if (Character.isDigit(c))
	    return true;
	if (IsOther(c))
	    return true;
	return false;
    }

    /**
     * turns the grammar into a string that can be saved in a file
     * precondition: start symbol on the left of first rule. if this is not
     * met the order of the rules will be changed.
     */
    public String toSaveString() throws Exception
    {
       String result = "";
       String left;
       String right;
       boolean found = false;
       Hashtable vars = new Hashtable();

       //map all of the variables into the hashtable
       String str;
       for(Enumeration e = myVariables.elements(); e.hasMoreElements(); )
       {
          str = (String) e.nextElement();
          
          if(str.length() == 1)
             vars.put(str, str);
       }

       //move the start symbol to the top of the grammar
       for(int pos = 0; pos < myProdsLeft.size(); ++pos)
       {
          left = (String) myProdsLeft.elementAt(pos);
          if(left.trim().equals(getStartSymbol().trim()))
          {
             myProdsLeft.removeElementAt(pos);
             myProdsLeft.insertElementAt(left, 0);

             right = (String) myProdsRight.elementAt(pos);
             myProdsRight.removeElementAt(pos);
             myProdsRight.insertElementAt(right, 0);

             found = true;
          }
       }
 
       if(!found)
         throw new Exception("The start symbol is not used.");
 
       //add all of the rules to the string
       int end;
       int strpos;
       String temp = "";
       char curVar = 'A';
       for(int pos = 0; pos < myProdsLeft.size(); ++pos)
       {
	   strpos = 0;
	   left = "";
	   while(strpos < ((String) myProdsLeft.elementAt(pos)).length())
	   {
               if(Character.isLowerCase(
                  ((String) myProdsLeft.elementAt(pos)).charAt(strpos)))
               {
                  left += "" + ((String) myProdsLeft.elementAt(pos)).charAt(strpos);
                  ++strpos;
                  continue;
               }
	       end = getEnd((String) myProdsLeft.elementAt(pos), strpos);
	       temp = ((String) myProdsLeft.elementAt(pos)).substring(strpos, end);
	       if(vars.get(temp) == null)
	       {
                  curVar = getNextVariable(curVar, vars);
		  vars.put(temp, "" + curVar);
                  left += "" + curVar;
	       }
               else
               {
                  left += (String) vars.get(temp);
               }
               strpos = end;
	   }

           strpos = 0;
	   right = "";
           while(strpos < ((String) myProdsRight.elementAt(pos)).length())
	   {
               if(Character.isLowerCase(
                  ((String) myProdsRight.elementAt(pos)).charAt(strpos)))
               {
                  right += "" + ((String) myProdsRight.elementAt(pos)).charAt(strpos);
                  ++strpos;
                  continue;
               }
	       end = getEnd((String) myProdsRight.elementAt(pos), strpos);
	       temp = ((String) myProdsRight.elementAt(pos)).substring(strpos, end);
	       if(vars.get(temp) == null)
	       {
                  curVar = getNextVariable(curVar, vars);
		  vars.put(temp, "" + curVar);
                  right += "" + curVar;
	       }
               else
               {
                  right += (String) vars.get(temp);
               }
               strpos = end;
	   }

	   if(right.equals("" + Params.lambdaChar))
	      right = "";
	   result += left + " " + arrow + " " + right + "\n";
       }

       if(curVar > 'Z')
       {
          throw new Exception("Too many variables.");
       }

       return result;
    }

    private char getNextVariable(char currentVar, Hashtable v)
    {
       boolean okay = false;
       String nextVar;

       nextVar = currentVar + "";
       boolean found = false;
       for(Enumeration e = v.elements(); e.hasMoreElements(); )
       {
         if(((String) e.nextElement()).trim().equals(nextVar))
         {
           found = true;
         }
       }
       if(!found)
       {
          return currentVar;
       }

       while(!okay)
       {
         nextVar = (++currentVar) + "";
         okay = true;
         for(Enumeration e = v.elements(); e.hasMoreElements(); )
         {
           if(((String) e.nextElement()).trim().equals(nextVar))
             okay = false;
         }
         if(currentVar == 'I')
	    okay = false;
       }
       return currentVar;
    }

    //private variables
    private Vector myVariables;
    private Vector myTerminals;
    private Vector myProdsLeft;
    private Vector myProdsRight;
    private int myNumVars;
    private int myNumTerms;
    private int myNumProds;
    private int myV;
    private int myT;
    private int myP;
    private int myArrowPosition;
}
