/*---------------------------------------------------------------------------
File:                 Parser.java
Package:              PateLib  Version 1.0
Author:               Anna O. Bilska (Anya) V1.0 08/07/96
Modified:             Ted Hung 8/99	
--------------------------------------------------------------------------*/

/* 
 * Susan H. Rodger, Anna O. Bilska (Anya), Jason R. Salemme
 * Computer Science Department
 * Duke University
 * August 1996
 * Supported by National Science Foundation DUE-9596002 and DUE-9555084.
 *
 * Copyright (c) 1996
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the author.  The name of the author may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

package PateLib;

import java.lang.*;
import java.util.*;

// The following class implements an exhaustive 
// search parser for grammars.
// It works as a thread; one must call
// SearchFor(someword) first, however, 
// then start the parser-thread.  
// The parser thread should run along a
// ParserMonitor thread, which checks when the parser
// finishes, and prints out the results.

public class Parser extends Thread
{
	public Parser(Grammar gramm,boolean flag)
	// constructor; expects a grammar, and
	// a flag which determines whether the 
	// parser ignores branches that already
	// exist
	{
		myG = gramm;
		ignoreduplicate = flag;
		myPath = new Vector();
		myProdsUsed = new Vector();
		isdoneParsing = false;
		found = false;
		isUnrestricted = false;
		isPaused = false;
		isStopped = false;
		iterations = 0;
		shiftee = new Shifter();
		AlreadyNodes = new PositionTree();
		myQ = new Queue();
		
	}

	public void SearchFor(String lookedFor)
	// postcondition: the parser knows what to
	// search for;
	{
		myLookedFor = new String(lookedFor);
	}


	public synchronized void CheckGrammarForUnrestricted()
	// postcondition: sets a flag to true if the grammar is
	// unrestricted or contains more than 1 character
	// on the left side of a production
	{
		for (myG.FirstProd(); !myG.ProdsDone(); myG.NextProd())
		{
				if ((myG.CurrentLeftProd()).length() > 1)
				{
						isUnrestricted = true;
						return;
				}
		}
	}


	public void run()
	// the run method which is executed when
	// the thread is start()ed
	{
		isdoneParsing = false;
		CheckGrammarForUnrestricted();		
		if(ignoreduplicate)
			CreateTree_Ignore(myLookedFor);
		else
			CreateTree(myLookedFor);
		isdoneParsing = true;
		SavePath();
		CleanUp();
		safeStop();
	}

	public boolean IsDoneParsing()
	// postcondition: returns true is the parser
	// has finished parsing
	{
		return isdoneParsing;
	}


	public boolean IsPaused()
	// postcondition: returns true if the parser
	// is paused
	{
		return isPaused;
	}
	
	public synchronized void PausePConstant()
	// postcondition:  the parser's pause constant
	// is set to true
	{
           isPaused = true;
	}

	public synchronized void UnPausePConstant()
	// postcondition: the parser's pause constant
	// is set to false
	{
           isPaused = false;
           this.notify();
	}

	public boolean IsFound()
	// postcondition: returns true if the word
	// was found in the language
	{
		return found;
	}

	public int GetIterations()
	// postcondition: returns how many nodes were
	// created in the tree
	{
           return iterations;
	}

	public Vector GetDerivation()
	// postcondition: returns the derivation
	// of the word, if the word was in the 
	// language
	{
           return myPath;
	}
	
	public Vector GetPathProds()
	// postcondition: returns the productions
	// that were used in the derivation
	{
		return myProdsUsed;
	}
	
	private synchronized void SavePath()
	// precondition: the word was found, myAt is
	// still pointing to the node which contains
	// the derived word
	// postcondition: the derivation is extracted
	// from the parser-tree-structure
	{
		if (found)
		{
			myPath.removeAllElements();
			myProdsUsed.removeAllElements();
			while(myAt != null)
			{
				myPath.addElement(myAt.data);
				myProdsUsed.addElement(new Integer(myAt.prod));
				myAt = myAt.parent;
			}
		}
	}

	
	
	public synchronized void CleanUp()
	// postcondition: the parser tree structure
	// is destroyed (derivation is not touched,
	// however)
	{
		myAt = myCurrentParent = null;
		myQ.MakeEmpty();
		AlreadyNodes = new PositionTree();
	}

	private synchronized boolean IsTerminal(char q)
	// postcondition: returns true if 'q' is a 
	// possible terminal (not necessarily in this
	// grammar)
	{
		return (myG.IsPossibleTerminal(q));
	}

	private synchronized boolean IsTerminal(String s)
	// postcondition: returns true if a string
	// is a possible terminal string (not necessarily 
	// in this grammar
	{
		for (int i=0; i<s.length(); i++)
		{
			if (!IsTerminal(s.charAt(i)))
				return false;
		}
		return true;
	}

	private synchronized int TerminalLength(String word)
	// postcondition: returns the length of the 
	// possible terminals in a string
	{
		int l=0;	
		for (int i=0; i<word.length(); i++)
		{
			if (IsTerminal(word.charAt(i)))
				l++;
		}
		return l;
	}


	
	public synchronized boolean Matches(String lookedfor, String neword)
	// postcondition: returns false if there is a terminal
	// string within 'neword' which is not in 'lookedfor'
	{
		String sub = new String();
		int i=0;
		int l=neword.length();
		boolean flag = false;
		while(i<l)
		{
				sub = new String();
				while((i<l) && IsTerminal(neword.charAt(i)))
				{
					sub += neword.charAt(i);
					i++;
					flag = true;
				}
				if (flag)
				{
					flag = false;
					i--;
				}
				if (lookedfor.indexOf(sub)==-1)
					return false;
				i++;
		}
		return true;
	}


	
	public synchronized boolean FrontEndMatches(String lookedfor, String neword)
	// postcondition:  return false if a terminal ending or beginning
	// of the 'neword' string does not match the ending or beginning
	// of 'lookedfor' 
	{
		int newordLen = neword.length();
		int lookedforLen = lookedfor.length();
		int i=0;
		int j;
		while ((i<newordLen) && (i<lookedforLen) && IsTerminal(neword.charAt(i)))
		{
			if (lookedfor.charAt(i) != neword.charAt(i))
				return false;
			i++;
		}
		
		i = newordLen-1;
		j = lookedforLen-1;
		while ((i>=0) && (j>=0) && IsTerminal(neword.charAt(i)))
		{
			if (neword.charAt(i) != lookedfor.charAt(j))
				return false;
			i--;
			j--;
		}
		return true;	
	}
	
	private synchronized boolean OKwith(String lookedfor, String neword)
	// postcondition: returns true if 'neword' could possibly
	// derive 'lookedfor' for a restricted grammar
	{
		if (!isUnrestricted)
		{
			if (TerminalLength(neword) > lookedfor.length()) 
			{
				return false;
			}
			if (!FrontEndMatches(lookedfor,neword))
			{
				return false;
			}
			if ( (IsTerminal(neword)) && (!neword.equals(lookedfor)))
			{	
				return false;
			}
			if (!Matches(lookedfor,neword))
			{
				return false;
			}
			
			return true;
		}
		return true;
	}

	public synchronized String TurnLambdaIntoSpace(String word)
	// precondition: assumes that there is only one
	// LAMDA symbol in 'word'
	// postcondition: deletes the first occurance of the 
	// LAMDA symbol from 'word'
	{
		int index = word.indexOf(Params.lambdaChar);
		
		if (index != -1)
		{
				String part1 = new String();
				String part2 = new String();
				part1 = word.substring(0,index);
				if (index == word.length())
					part2 ="";
				else
					part2 = word.substring(index+1,word.length());
				return(part1 + part2);
		}
		return word;
	}

	public synchronized Vector GetPossibilities_Ignore(String c)
	// postconditon: returns a vector with all the possible
	// derivations from 'c' based on the productions in the 
	// grammar;  DOES NOT allow already created (in the tree)
	// strings to be placed in it again 
	{
		int temp;
		GString myCurrentWord = new GString(c);
		myPossibilities.removeAllElements();
		myPositions.removeAllElements();
		myProdsUsed.removeAllElements();
		int k=0;

		for (myG.FirstProd(); !myG.ProdsDone(); myG.NextProd())
		{
			if (myCurrentWord.Contains(myG.CurrentLeftProd(), myPositions)) 
			{
				for (int i=0; i< myPositions.size(); i++)
				{	
					myIndex = ((Integer)myPositions.elementAt(i)).intValue();	
					newWord = myCurrentWord.Exchange(myIndex,myG.CurrentLeftProd(),
						  myG.CurrentRightProd());	
					newWord = TurnLambdaIntoSpace(newWord);
					if (OKwith(myLookedFor,newWord) && 
						(!AlreadyNodes.ContainsAlsoAdd(newWord)))
					{
						myPossibilities.addElement(newWord);
						temp = shiftee.TwoIntoOne(myIndex,k);
						myProdsUsed.addElement(new Integer(temp));
						iterations++;
				        }
				}
			}	
			k++;
					
		}
		return myPossibilities;
	}

	public synchronized Vector GetPossibilities(String c)
	// postconditon: returns a vector with all the possible
	// derivations from 'c' based on the productions in the 
	// grammar;  allows already created (in the tree)
	// strings to be placed in it again 
	{
		int temp;
		GString myCurrentWord = new GString(c);
		myPossibilities.removeAllElements();
		myPositions.removeAllElements();
		myProdsUsed.removeAllElements();
		int k=0;

		for (myG.FirstProd(); !myG.ProdsDone(); myG.NextProd())
		{
			if (myCurrentWord.Contains(myG.CurrentLeftProd(), myPositions)) 
			{
				for (int i=0; i< myPositions.size(); i++)
				{	
					myIndex = ((Integer)myPositions.elementAt(i)).intValue();	
					newWord = myCurrentWord.Exchange(myIndex,myG.CurrentLeftProd(),
								myG.CurrentRightProd());	
					newWord = TurnLambdaIntoSpace(newWord);
					if (OKwith(myLookedFor,newWord))
					{
						iterations++;
						myPossibilities.addElement(newWord);
						temp = shiftee.TwoIntoOne(myIndex,k);
						myProdsUsed.addElement(new Integer(temp));
				        }
				}
			}	
			k++;
					
		}
		return myPossibilities;
	}	

	private void CreateTree(String lookedFor)
	// postcondition:  creates a derivation tree for 
	// a word 'lookedfor'
	{
		myQ.MakeEmpty();
		iterations = 0;
		TreeLink myCurrentParent = new TreeLink("S",0,null,null,null);
		AlreadyNodes = new PositionTree();
		myQ.Enqueue(myCurrentParent);
		found = false;

		while (((!myQ.IsEmpty()) && (!found)))
		{
			myCurrentParent = (TreeLink)myQ.Dequeue();
		
			try 
			{
                           sleep(4);
                           synchronized(this)
                           {
                              while (isPaused)
                                 this.wait();
                              if(isStopped)
                                 return;
                           }
			}
			catch(InterruptedException e)
			{
			}

			myPossibilities = GetPossibilities((String)myCurrentParent.data);
			myAt = myCurrentParent;
			
			int i=0;
			if ((!myPossibilities.isEmpty()) && (!found))
			{	
				myAt.firstchild = new TreeLink(myPossibilities.elementAt(i),
						((Integer)myProdsUsed.elementAt(i)).intValue(),
						null,null,myCurrentParent);
				myAt = myAt.firstchild;
				myQ.Enqueue(myAt);
				if (((String)myPossibilities.elementAt(i)).equals(lookedFor))
				{
					found = true;
			 	}
				i++;
			}
			while ((i<myPossibilities.size()) && (!found))
			{
				myAt.sibling = new TreeLink(myPossibilities.elementAt(i),
						((Integer)myProdsUsed.elementAt(i)).intValue(),
						null,null, myCurrentParent);
				myAt = myAt.sibling;
				myQ.Enqueue(myAt);
				if (((String)myPossibilities.elementAt(i)).equals(lookedFor))
				{
					found = true;
				}
				i++;
			}
		}	
	}

	private void CreateTree_Ignore(String lookedFor)
	// postcondition: creates a derivation tree for 
	// a word 'lookedFor' -- DOES NOT allow duplicate 
	// strings to be placed in the tree
	{
		myQ.MakeEmpty();
		iterations = 0;
		TreeLink myCurrentParent = new TreeLink("S",0,null,null,null);
		myQ.Enqueue(myCurrentParent);
		found = false;
	
		while (((!myQ.IsEmpty()) && (!found)))
		{
			myCurrentParent = (TreeLink)myQ.Dequeue();
			try 
			{
                           sleep(4);
                           synchronized(this)
                           {
                              while (isPaused)
                                 this.wait();
                              if(isStopped)
                                 return;
                           }
			}
			catch(InterruptedException e)
			{
			}
			myPossibilities = GetPossibilities_Ignore((String)myCurrentParent.data);
			myAt = myCurrentParent;
			
			int i=0;

			// add a firstchild to the parent
			if ((!myPossibilities.isEmpty()) && (!found))
			{
				myAt.firstchild = new TreeLink(myPossibilities.elementAt(i),
						((Integer)myProdsUsed.elementAt(i)).intValue(),
						null,null,myCurrentParent);
				myAt = myAt.firstchild;
				myQ.Enqueue(myAt);
				if (((String)myPossibilities.elementAt(i)).equals(lookedFor))
				{
					found = true;
			 	}
				i++;
			}
			// add siblings to the firstchild
			while ((i<myPossibilities.size()) && (!found))
			{
				myAt.sibling = new TreeLink(myPossibilities.elementAt(i),
						((Integer)myProdsUsed.elementAt(i)).intValue(),
						null,null, myCurrentParent);
				myAt = myAt.sibling;
				myQ.Enqueue(myAt);
				if (((String)myPossibilities.elementAt(i)).equals(lookedFor))
				{
					found = true;
				}
				i++;
			}
		}
	
	}

	public void safeStop()
	{
	    isStopped = true;
	}

	public boolean isDone()
	{
	    return isStopped;
	}

	private TreeLink myAt;
	private TreeLink myCurrentParent;
	private Vector myPositions = new Vector();
	private Vector myPossibilities = new Vector();
	private Vector myPath;
	private String newWord = new String();
	private Grammar myG = new Grammar();
	private int myIndex;
	private boolean found;
	private Queue myQ;
	private String myLookedFor;
	private Vector myProdsUsed;
	private PositionTree AlreadyNodes;
	private boolean ignoreduplicate = true;
	private int iterations;
	private boolean isdoneParsing;
	private boolean isUnrestricted;
	private Shifter shiftee;
	private boolean isPaused;
	private boolean isStopped;
}


class TreeLink
// a node of the tree created by the parser
{
	Object data;             // stores the actual string
	TreeLink sibling;        // pointer to sibling 
	TreeLink firstchild;	 // pointer to firstchild	
	TreeLink parent;         // pointer to parent
	int prod;				 // stores which production was
							 // used to create this specific 
							 // word and the position where
							 // within the string

	TreeLink(Object o, int d,TreeLink s, TreeLink f, TreeLink p)
	{
		data=o;
		sibling=s;
		firstchild=f;
		parent=p;
		prod=d;
	}

}


